# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pacu',
 'pacu.core',
 'pacu.core.secretfinder',
 'pacu.modules',
 'pacu.modules.acm__enum',
 'pacu.modules.api_gateway__create_api_keys',
 'pacu.modules.apigateway__enum',
 'pacu.modules.aws__enum_account',
 'pacu.modules.aws__enum_spend',
 'pacu.modules.cloudformation__download_data',
 'pacu.modules.cloudtrail__csv_injection',
 'pacu.modules.cloudtrail__download_event_history',
 'pacu.modules.cloudwatch__download_logs',
 'pacu.modules.codebuild__enum',
 'pacu.modules.detection__disruption',
 'pacu.modules.detection__enum_services',
 'pacu.modules.dynamodb__enum',
 'pacu.modules.ebs__download_snapshots',
 'pacu.modules.ebs__enum_volumes_snapshots',
 'pacu.modules.ebs__explore_snapshots',
 'pacu.modules.ec2__backdoor_ec2_sec_groups',
 'pacu.modules.ec2__check_termination_protection',
 'pacu.modules.ec2__download_userdata',
 'pacu.modules.ec2__enum',
 'pacu.modules.ec2__startup_shell_script',
 'pacu.modules.ecr__enum',
 'pacu.modules.ecs__backdoor_task_def',
 'pacu.modules.ecs__enum',
 'pacu.modules.ecs__enum_task_def',
 'pacu.modules.elb__enum_logging',
 'pacu.modules.enum__secrets',
 'pacu.modules.glue__enum',
 'pacu.modules.guardduty__list_accounts',
 'pacu.modules.guardduty__list_findings',
 'pacu.modules.guardduty__whitelist_ip',
 'pacu.modules.iam__backdoor_assume_role',
 'pacu.modules.iam__backdoor_users_keys',
 'pacu.modules.iam__backdoor_users_password',
 'pacu.modules.iam__bruteforce_permissions',
 'pacu.modules.iam__detect_honeytokens',
 'pacu.modules.iam__enum_permissions',
 'pacu.modules.iam__enum_roles',
 'pacu.modules.iam__enum_users',
 'pacu.modules.iam__enum_users_roles_policies_groups',
 'pacu.modules.iam__get_credential_report',
 'pacu.modules.iam__privesc_scan',
 'pacu.modules.inspector__get_reports',
 'pacu.modules.lambda__backdoor_new_roles',
 'pacu.modules.lambda__backdoor_new_sec_groups',
 'pacu.modules.lambda__backdoor_new_users',
 'pacu.modules.lambda__enum',
 'pacu.modules.lightsail__download_ssh_keys',
 'pacu.modules.lightsail__enum',
 'pacu.modules.lightsail__generate_ssh_keys',
 'pacu.modules.lightsail__generate_temp_access',
 'pacu.modules.rds__explore_snapshots',
 'pacu.modules.route53__enum',
 'pacu.modules.s3__download_bucket',
 'pacu.modules.systemsmanager__download_parameters',
 'pacu.modules.systemsmanager__rce_ec2',
 'pacu.modules.vpc__enum_lateral_movement',
 'pacu.modules.waf__enum',
 'pacu.modules_archive.iam__enum_assume_role',
 'pacu.modules_archive.s3__bucket_finder']

package_data = \
{'': ['*'],
 'pacu.modules.iam__privesc_scan': ['PassExistingRoleToNewCodeStarProject/*']}

install_requires = \
['SQLAlchemy-Utils>=0.36.8,<0.37.0',
 'SQLAlchemy>=1.3.0,<1.4.0',
 'awscli>=1.18,<2.0',
 'boto3>=1.16,<2.0',
 'botocore>=1.16,<2.0',
 'dsnap>=1.0.0,<2.0.0',
 'requests>=2.25.1,<3.0.0',
 'typing-extensions>=3.7.4.3,<4.0.0.0',
 'urllib3>=1.26.4,<2.0.0']

entry_points = \
{'console_scripts': ['pacu = pacu.__main__:main']}

setup_kwargs = {
    'name': 'pacu',
    'version': '0.1.15',
    'description': '',
    'long_description': None,
    'author': 'Ryan Gerstenkorn',
    'author_email': 'ryan.gerstenkorn@rhinosecuritylabs.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
